"""
Utility functions built on spaCy.
"""

import spacy
import re

"""
Load the spaCy English language model one time for the entire application.
"""
spacyEnglishModel = spacy.load('en_core_web_lg')
lightSpacyEnglishModel = spacy.load('en_core_web_lg', 
                                    disable=["tagger", "parser", "ner", "textcat", "lemmatizer"])

def tokenize(inputStr):
    """
    Use the lightweight English model to tokenize a piece of text.
    Fixes a couple of bugs in the default Spacy tokenizer.
    :param inputStr: 
    :return: a list of tokens
    """
    
    inputStr = inputStr.strip()

    # Bug 1: Hyphen issue: if the input consists solely of '- + word' or 'word + -',
    # Spacy fails to separate the hyphen from the word.
    if inputStr.startswith('-'):
        inputStr = re.sub('-(\\w)', '- \\1', inputStr)
    if inputStr.endswith('-'):
        inputStr = re.sub('(\\w)-', '\\1 -', inputStr)

    # Tokenize with a light-weight Spacy doc.
    lightweightDoc = lightSpacyEnglishModel(inputStr)

    # Bug 2: Spacy outputs strings of whitespace as tokens. Strip these out here.)
    return [str(x) for x in lightweightDoc if str(x).strip() != '']


if __name__ == '__main__':
    pass



import re

from typing import List

from text_to_relations.relation_extraction.ExtractionPhaseABC import ExtractionPhaseABC
from text_to_relations.relation_extraction.RegexString import RegexString
from text_to_relations.relation_extraction.Annotation import Annotation
from text_to_relations.relation_extraction.TokenAnn import TokenAnn

class MinMaxPhase_1(ExtractionPhaseABC):

    def __init__(self, doc_contents:str, 
                 given_annotations: List[Annotation],
                 verbose: bool=False):
        """
        Identify phrases like 'between 170 and 220 pounds' and 
            'within the range of 60 to 90 points' by looking for:
            RangeMarker + Number + Number + Unit_of_Measure
        Args:
            doc_contents (str): normalized contents of the document
                being processed
            given_annotations (List[Annotation]): list of annotations/
                entities which are going to be used to form new
                relations
            verbose (bool, optional): Defaults to False.
        """
        super().__init__(doc_contents)

        required_annotation_types = ['Number', 'Unit_of_Measure']
        self.given_annotations = []
        for ann in given_annotations:
            if ann.type not in required_annotation_types:
                continue
            self.given_annotations.append(ann)
        self.verbose = verbose

        if self.verbose:
            print(f"self.given_annotations: {self.given_annotations}")
        

    def run_phase(self) -> List[Annotation]:
        """
        Find MinMax entities for the given document and annotations.
        Returns:
            List[Annotation]: list of annotations whose type is 'MinMax'
        """
        # Use RegexString to create our own annotations (entities).
        range_markers = ['within the range of', 'within', 'between']
        regex_str = RegexString(range_markers)

        range_anns = []
        triples = regex_str.get_match_triples(self.doc_contents)
        for triple in triples:
            ann = Annotation('RangeMarker', triple[0], triple[1], triple[2])
            range_anns.append(ann)

        if self.verbose: print(f"range_anns: {range_anns}")

        # Sort all annotations by offset.
        anns = self.given_annotations
        anns.extend(range_anns)
        anns = Annotation.sort(anns)
        if self.verbose: print(f"\nsorted anns: {anns}\n")
        
        # Replace the text document with a new representation in which every 
        # token is an element--except those tokens which are overlapped by
        # one of our previously-created annotations/entities.
        annotation_view_str = ExtractionPhaseABC.build_merged_representation(self.doc_contents, anns)

        # Display each token or annotation on its own line.
        if self.verbose:
            ann_list = ExtractionPhaseABC.merged_representation_to_Annotations(annotation_view_str)
            for ann in ann_list:
                print(f"ann: {ann}")

        new_annotations = self.check_annotation_proximity(annotation_view_str=annotation_view_str)
        return new_annotations

    def check_annotation_proximity(self, annotation_view_str: List[Annotation]) -> List[Annotation]:
        """
        Look for parts of the text where the annotations we are interested
        in are in close proximity.
        
        Args:
            annotation_view_str (List[Annotation]): the annotation-only representation of the 
                input document

        Returns:
            List[Annotation]: a list of any new annotations created
        """
        if self.verbose: print(f"\nEntering check_annotation_proximity( ).")

        new_annotations = []

        # Now we look for parts of the text where the annotations we are interested
        # in are in close proximity.
        regex_1 = TokenAnn.build_annotation_distance_regex("RangeMarker", (0, 3), None, "Number")

        # Create a list of (substring, start_offset, end_offset) triples for RangeMarker + Cardinal.
        match_1_triples = [(m.group(), m.start(), m.end()) for m in re.finditer(regex_1, annotation_view_str)]

        # Triple-nested for loop which tests for four annotations in the correct order and
        # in close proximity in order to form a MinMax relation. An annotation can belong
        # to only one relation, and no relations overlap.
        # This means that if a relation is created, we want to break from both Loop 3
        # and Loop 2, and continue on Loop 1. We implement this behavior using a
        # `breaking` flag. 
        # Moreover, for the sake of efficiency, when we continue with the outer loop
        # we don't want the restarted inner loops to iterate over annotations that 
        # precede the end of the previous loop's match. We implement this with checks at 
        # the start of the inner loops, e.g. "if m1_trip[1] > m2_trip[1]:"
        for m1_trip in match_1_triples:
            breaking = False
            if self.verbose: print(f"\nFound first match. m1_trip: {m1_trip}")

            regex_2 = TokenAnn.build_annotation_distance_regex("Number", (0, 2), None, "Number")
            # Create a list of (substring, start_offset, end_offset) triples for Cardinal + Cardinal.
            match_2_triples = [(m.group(), m.start(), m.end()) for m in re.finditer(regex_2, annotation_view_str)]

            for m2_trip in match_2_triples:
                if breaking is True:
                    break
                if self.verbose: print(f"\n  Found second match. m2_trip: {m2_trip}")

                if m1_trip[1] > m2_trip[1]:
                    continue

                regex_3 = TokenAnn.build_annotation_distance_regex("Number", (0, 2), None, "Unit_of_Measure")
                # Create a list of (substring, start_offset, end_offset) triples for Cardinal + UoM.
                match_3_triples = [(m.group(), m.start(), m.end()) for m in re.finditer(regex_3, annotation_view_str)]

                for m3_trip in match_3_triples:
                    if self.verbose: print(f"\n    Found third match. m3_trip: {m3_trip}")

                    if m2_trip[1] > m3_trip[1]:
                        continue
                    # Create the new relation, which spans from the start of the RangeMarker to the
                    # end of the Unit_of_Measurement.

                    # Get the starting position of the RangeMarker in the first match.
                    # Ignore [1] and [2], the offsets of the RangeMarker-Cardinal 
                    # match in the merged representation.
                    text_matched = m1_trip[0]   
                    m1_anns = ExtractionPhaseABC.merged_representation_to_Annotations(text_matched)
                    start = m1_anns[0].start_offset

                    # Get the ending position of the last annotation in the third match.
                    text_matched = m3_trip[0]   # Ignore [1] and [2], the offsets.
                    m3_anns = ExtractionPhaseABC.merged_representation_to_Annotations(text_matched)
                    end = m3_anns[1].end_offset

                    substr = self.doc_contents[start:end]

                    # Finally we have everything we need for our new annotation.
                    min_max = Annotation('MinMax', substr, start, end)

                    if self.verbose: print(f"New annotation created: {min_max}")

                    new_annotations.append(min_max)
                    breaking = True
                    break

        if self.verbose:
            print("\nNew annotations:")
            for new_ann in new_annotations:
                print(f"  {new_ann}")
            print()

        return new_annotations



from typing import List, Self

from text_to_relations.relation_extraction.Annotation import Annotation
import spacy

spacy_model = spacy.load('en_core_web_lg', disable=["tagger", "ner", "lemmatizer"])
spacy_model.add_pipe('sentencizer')

class SentenceAnn(Annotation):
    """
    An Annotation object covering an entire sentence in a document
    """
    
    def __init__(self, contents: str, start_offset: int, end_offset: int):
        """
        Args:
            contents (str): sentence text
            start_offset (int): start offset of the sentence in the doc
            end_offset (int): end offset in the doc
        """
        super().__init__('Sentence', contents, start_offset, end_offset)
    
    
    @staticmethod
    def text_to_SentenceAnns(input: str) -> List[Self]:
        """
        Split the given input text into sentences, and create a SentenceAnn
        on each one.
        Args:
            input (str): the text to split

        Returns:
            List[Self]: a list of SentenceAnn annotations created on the 
                given text
        """
        sentenceSpans = spacy_model(input).sents
        sentenceStrs = []
        for sentence in sentenceSpans:
            sentenceStrs.append(sentence.text.strip())
        
        result = []
        startSearchIdx = 0
        for sentence in sentenceStrs:
            startIdx = input.find(sentence, startSearchIdx)
            endIdx = startIdx + len(sentence)
            sentAnn = SentenceAnn(sentence.strip(), startIdx, endIdx)
            result.append(sentAnn)
            startSearchIdx = endIdx

        return result
    

if __name__ == '__main__':
    pass



import re

import unicodedata

regexMultipleSpaces = re.compile(r' +', re.IGNORECASE)

def removeMultipleSpaces(inStr):
    """ Remove multiple consecutive spaces from a string. """
    return regexMultipleSpaces.sub(' ', inStr)


def isAllPunc(inputString):
    """
    Is the given string all punctuation?
    :param inputString:
    :return: boolean
    """
    bAllPunc = True
    for char in inputString:
        category = unicodedata.category(char)
        if category.startswith('P') or category == 'Sm':
            continue
        bAllPunc = False
        break

    return bAllPunc


def isAllWordChars(inputStr):
    """
    Does the given string consist only of word characters?
    :return:
    """
    regexWordChar = re.compile(r'^\w+$')

    if re.match(regexWordChar, inputStr):
        return True
    return False



"""
Abstract base class for relation extraction--i.e., building
relations between previously-identified entities.
"""
import re
from abc import ABCMeta, abstractmethod
from typing import List

from text_to_relations.relation_extraction.TokenAnn import TokenAnn
from text_to_relations.relation_extraction.Annotation import Annotation

class ExtractionPhaseABC(metaclass=ABCMeta):
    """
    Abstract base class of phases.
    """
    regexWhitespace = re.compile(r'\s+', re.IGNORECASE | re.DOTALL | re.MULTILINE)

    @abstractmethod
    def run_phase(self):
        """
        Runs this extraction phase object.
        """
        pass

    def __init__(self, doc_contents: str):
        """

        Args:
            doc_contents (str): the normalized contents of the 
                document being processed
        """
        if doc_contents is None or doc_contents == '':
            msg = "doc_contents is empty or None. "
            msg += "An extraction phase object requires a document to process."
            raise TypeError(msg)
        self.doc_contents = doc_contents

    @staticmethod
    def build_merged_representation(doc_contents: str, 
                                    anns: List[Annotation], 
                                    verbose: bool=False) -> str:
        """
        Create an Annotation-only representation of the document by
        merging the given bespoke annotations on it into a TokenAnn list 
        for all the other tokens in the doc.
        Args:
            doc_contents (str): the normalized contents of the doc being 
                processed
            anns (List[Annotation]): a list of bespoke annotations you want 
                to appear merged into the doc
            verbose (bool, optional): Defaults to False.

        Raises:
            ValueError: If the process fails to insert any of the provided 
                bespoke annotations into the final result

        Returns:
            str: a string representing all the annotations in the document,
                sorted by offset
        """
        contents = doc_contents.rstrip()

        # If this is empty after the process, something may be wrong.
        unconsumedAnnotations = anns

        result = ""
        lastPos = 0
        
        # Strategy: Tokenize the doc and iterate through all the tokens. If a token
        # is covered by an annotation, write that annotation to the output and advance
        # lastPos to the end of the annotation; otherwise, write the token to output
        # and continue.
        tokensObjs = TokenAnn.get_token_objects(contents, 0)
        
        for tokenObj in tokensObjs:
            
            if lastPos > tokenObj.start_offset:
                continue

            tempAnns = unconsumedAnnotations
            foundAnn = False
            for ann in tempAnns:
                if ann.start_offset <= tokenObj.start_offset:
                    # Write the annotation and remove it from the unconsumedAnnotations.
                    unconsumedAnnotations = unconsumedAnnotations[1:]
                    result += str(ann)
                    if verbose:
                        print(str(ann))
                    lastPos = ann.end_offset
                    foundAnn = True
                else:
                    break

            if foundAnn:
                continue
            
            # This token occurs in the document before the next unconsumed annotation. Write it to 
            # output.
            result += str(tokenObj)

            lastPos = tokenObj.end_offset

        # Verify that all the annotations have been consumed.
        if len(unconsumedAnnotations) > 0:
            msg = "Final annotations in the anns parameter not inserted into the merged document."
            msg += f"  uncomsumed annotations: {unconsumedAnnotations}"
            raise ValueError(msg)

        return result


    @staticmethod
    def merged_representation_to_Annotations(rep: str, 
                                    verbose: bool=False) -> List[Annotation]:
        """
        Essentially reverses build_merged_representation(). From a merged representation,
        or a subset thereof, create a list of Annotations.
        Args:
            rep (str): merged representation, e.g. a string looking like this
                (ignore the backslashes): 
                    <'CARDINAL'(normalizedContents='80', start='92', end='94')> \
                    <'Token'(normalizedContents='to', start='95', end='97', kind='word')> \
                    <'CARDINAL'(normalizedContents='90', start='98', end='100')>
            verbose (bool, optional): Defaults to False.

        Returns:
            List[Annotation]: 
        """
        # Each ann is contained within angle brackets--<>.
        incomplete_strs = rep.split('<')
        complete_strs = []
        for in_str in incomplete_strs:
            if not in_str:
                continue
            if verbose: print(f"in_str: {in_str}")
            complete_strs.append('<' + in_str)
        if verbose: print(f"complete_strs: {complete_strs}")
        anns = [Annotation.str_to_Annotation(c_str) for c_str in complete_strs]
        return anns



"""
Convert entity information into specific pre-determined relations
between those entities.
"""

from typing import Dict, List
from pathlib import Path
import sys

import inspect

# All project imports must occur after we have updated the path.
path = Path(__file__).absolute()
path = path.parent.parent.parent
print(f"Appending to system path: {str(path)}")
sys.path.append(str(path))

from text_to_relations.relation_extraction.RegexString import RegexString
from text_to_relations.relation_extraction.Annotation import Annotation
from text_to_relations.relation_extraction.min_max_phase_1 import MinMaxPhase_1
from text_to_relations.relation_extraction.min_max_phase_2 import MinMaxPhase_2


def run_extraction_phases(input: str, 
                          anns: List[Annotation],
                          verbose: bool=False) -> List[Annotation]:
    if verbose: print(f"\nEntering run_extraction_phases(). anns: {anns}\n")
    extraction_phase_1 = MinMaxPhase_1(input, anns, verbose=verbose)
    minmax_anns_1 = extraction_phase_1.run_phase()
    if verbose: 
        msg = f"\nReturned to run_extraction_phases(). minmax_anns: {minmax_anns_1}"
        print(msg)

    # Remove from the input annotations any which form part of the new
    # MinMax annotations. This prevents the next set of MinMax rules from
    # matching on those annotations.
    enclosed = []
    for mm_ann in minmax_anns_1:
        enclosed.extend(Annotation.get_enclosed(mm_ann, anns))
    anns_2 = [x for x in anns if x not in enclosed]
    
    # After this, anns_2 will contain the newly-found MinMax annotations
    # plus any unconsumed Number and Unit_of_Measure annotations.
    anns_2.extend(minmax_anns_1)
    if verbose: print(f"  Filtered anns_2 annotations: {anns}\n")

    extraction_phase_2 = MinMaxPhase_2(input, anns_2, verbose=verbose)
    minmax_anns_2 = extraction_phase_2.run_phase()
    result = minmax_anns_1 + minmax_anns_2

    return Annotation.sort(result)


def entities_to_annotations(entities: List[Dict[str, str]]) -> List[Annotation]:
    annotations = []
    for entity in entities:
        type = entity['type']
        start = entity['start']
        end = entity['end']
        text = entity['text']
        ann = Annotation(type, text, start, end)

        annotations.append(ann)

    return annotations

def entities_to_relations(input_dict: Dict[str, object],
                          verbose: bool=False) -> List[Dict[str, str]]:
    input_text = input_dict["text"]

    annotations = entities_to_annotations(input_dict["entities"])
    if verbose: print(f"Annotations created: {annotations}")
    
    new_anns = run_extraction_phases(input_text, annotations, verbose=verbose)

    if verbose:
        print(f"\nannotations: {annotations}")
        print(f"new_anns: {new_anns}\n")

    result = [x.to_dict() for x in new_anns]
    
    return result


if __name__ == '__main__':

    text = \
    """
    During those fraught times his weight ranged between 170 and 220 pounds
    and, with the 30 to 40 drinks per week he was inclined to enjoy, his IQ 
    varied almost as much--anywhere within the range of 60 to 90 points. 
    """
    input = inspect.cleandoc(text)

    entities = []

    # Write rules for two types of entities--Number and Unit_of_Measurement.
    number_rs = RegexString.regex_to_RegexString('(\d+)')
    matches = number_rs.get_match_triples(text)
    print("\nNumbers and offsets:")
    for match in matches:
        entity_dict = {
            'text': match[0],
            'start': match[1],
            'end': match[2],
            'type': 'Number'
        }
        entities.append(entity_dict)
        print(entity_dict)

    uom_rs = RegexString(['pounds', 'drinks', 'points'])
    matches = uom_rs.get_match_triples(text)
    print("\nUnits of measure and offsets:")
    for match in matches:
        entity_dict = {
            'text': match[0],
            'start': match[1],
            'end': match[2],
            'type': 'Unit_of_Measure'
        }
        entities.append(entity_dict)
        print(entity_dict)

    input_dict = {
        "text": text,
        "entities": entities
        }

    print()

    relations = entities_to_relations(input_dict, verbose=True)

    print(f"\nPrinting final output...")
    print(relations)



import re

from typing import List, Tuple, Self

class RegexString(object):
    """
    A class wrapped around a regular expression, offering functionality
    to create, concatenate and construct complex regex strings via an
    easy-to-use set of functions.
    """

    def __init__(self, match_strs: List[str], 
                 whole_word: bool=False,
                 optional: bool=False, 
                 non_capturing: bool=True,
                 prepend: str='', append: str=''):
        """

        Args:
            match_strs (List[str]): a collection of strings; if there 
                is more than one element, each element is to be OR'd together 
                for a match; if only one element, then there is no OR'd group
            whole_word (bool, optional): whether or not to require a match on
                whole words only. Defaults to True.
            optional (bool, optional): whether or not the OR'd items in the 
                regex should be optional. Defaults to False.
            non_capturing (bool, optional): if True, the OR'd items should 
                form a non-capturing group (i.e., '?:' should be inserted 
                after the left parenthesis). Note that other RegexString functions,
                such as get_match_triples() may behave unexpectedly if this
                is set to False. Defaults to True.
            prepend (str, optional): regex string to prepend to the OR'd list. 
                Defaults to ''.
            append (str, optional): regex string to append to the OR'd list. 
                Defaults to ''.
        """
        # If match_strs is a string, the user has made an error.
        if isinstance(match_strs, str):
            raise ValueError("match_strs parameter must be a list. You passed in a string.")

        # If passed a list of regex's re.findall() will take the first match found
        # even if a longer match is possible. E.g. if `text` contains "18" and we call
        # re.findall("(1|18)", text), only "1" will be found. Here we fix this
        # by sorting on string length descending on all strings, not bothering to
        # first check whether they are substrings.
        match_strs.sort(key=len, reverse=True)
        self.match_strs = match_strs

        self.whole_word = whole_word
        self.optional = optional
        self.non_capturing = non_capturing

        # '\b' causes debugging issues because in print statements it 
        # is invisible because it is interpreted as backspace. We don't 
        # attempt to fix this if '\b' is just a substring of a longer
        # prepend or append.
        esc_backslash_b = r'\b'
        if prepend == '\b':
            prepend = esc_backslash_b
        if append == '\b':
            append = esc_backslash_b
        self.prepend = prepend
        self.append = append

        if self.whole_word:
            msg = "Do not use 'whole_word=True' with '\\b' in "
            if r'\b' in self.prepend:
                msg += f"'prepend'. prepend = '{self.prepend}'."
                raise ValueError(msg)
            elif r'\b' in self.append:
                msg += f"'append'. append = '{self.append}'."
                raise ValueError(msg)

        self.regex_str = self.set_regex()


    def set_regex(self):
        result = ''

        if self.prepend != '':
            result += self.prepend

        if len(self.match_strs) > 1:
            # OR the match_strs together
            result += '('
            if self.non_capturing:
                result += '?:'

            for item in self.match_strs:
                result += item + '|'
            # Remove last pipe.
            result = result[0:-1]

            result += ')'

            if self.optional:
                result += '?'
        else:
            # There is only one item to put into the regex.
            if self.optional:
                # Make the item optional.
                result += '('
                if self.non_capturing:
                    result += '?:'
                result += self.match_strs[0]
                result += ')?'
            else:
                # The item is required.
                result += self.match_strs[0]
                if not self.non_capturing:
                    result = '(' + result + ')'

        if self.whole_word:
            result = r'\b' + result + r'\b'

        if self.append != '':
            result += self.append

        return result

    def get_regex_str(self):
        """
        Return the regular expression for this object's properties.
        :return:
        """
        return self.regex_str
    
    def get_match_triples(self, input: str) -> List[Tuple]:
        """
        Run re.finditer() on this regex.
        Note that this function is likely to fail if you have created any RegexString objects where non-group capturing is False.
        
        Args:
            input (str): text to run re.finditer() against.
        Returns:
            List[Tuple]: a list of (text-matched, start-offset, end-offset)
            triples.
        """
        match_triples = [(m.group(), m.start(), m.end()) for m in re.finditer(self.get_regex_str(), input)]
        return match_triples

    @staticmethod
    def concat(rs1: Self, 
               rs2: Self, 
               insert_opt_ws: bool=False) -> Self:
        """
        Create a new RegexString by concatenating the two input RegexString objects.
        Note: This method does not allow words between the two regex's. To
        allow words, use concat_with_word_distances().

        Args:
            rs1 (RegexString): 
            rs2 (RegexString): 
            insert_opt_ws (bool, optional): if True, allow a single whitespace 
            character between the regex expressions. Defaults to False.

        Raises:
            ValueError: if parameters are not RegexStings
        Returns:
            RegexString: the new RegexString
        """
        if not isinstance(rs1, RegexString):
            raise ValueError("rs1 parameter must be a RegexString object.")
        if not isinstance(rs2, RegexString):
            raise ValueError("rs2 parameter must be a RegexString object.")

        if insert_opt_ws:
            joinStrRegex = r'(?:\s)?'
        else:
            joinStrRegex = ''  # In other words, no whitespace allowed.

        # Create an empty/invalid RegexString object, and make it valid by editing
        # its regex_str property.
        resultRegexString = RegexString([''])
        resultRegexString.regex_str = rs1.regex_str + joinStrRegex + rs2.regex_str

        if rs2.optional:
            resultRegexString.optional = True
        else:
            resultRegexString.optional = False

        return resultRegexString


    @staticmethod
    def concat_with_word_distances(rs1: Self, 
                                   rs2: Self,  
                                   min_nbr_words: int=0, 
                                   max_nbr_words: int=0) -> Self:
        """
        Create a new RegexString by concatenating the two input 
        RegexString objects with a minimum and maximum number of 
        non-whitespace-character strings between them. Note that, 
        in this implementation, all of these strings are considered 
        single words: 'John', 'John:', '("John.")'.
        Note:
        - This functionality assumes that the input text does not have 
        spurious or accidental consecutive whitespace. In other words, if 
        the input text has 'a  dog' and you are looking for 'a' followed by 
        'dog', no match will occur because of the double spaces.

        Args:
            rs1 (RegexString): _description_
            rs2 (RegexString): _description_
            min_nbr_words (int, optional): the minimum number of words 
                allowed between the 
            two regex expressions. Defaults to 0.
            max_nbr_words (int, optional): the maximum number of words 
                allowed between the two regex expressions. Defaults to 0.
        Raises:
            ValueError: if parameters are not RegexStings
            ValueError: if min_nbr_words > max_nbr_words
        Returns:
            RegexString: the new RegexString
        """
        # Constants.
        wordRegex = r'\s\S+'
        readyForDistanceRangeWordRegex = r'(?:' + wordRegex + ')'

        interveningPunc = r'(?:\b\S+)?'
        interveningPuncAndSpace = interveningPunc + '\s'

        # Check input parameters.
        if not isinstance(rs1, RegexString):
            raise ValueError("rs1 parameter must be a RegexString object.")
        if not isinstance(rs2, RegexString):
            raise ValueError("rs2 parameter must be a RegexString object.")

        if min_nbr_words > max_nbr_words:
            raise ValueError("min_nbr_words cannot be greater than max_nbr_words.")

        # Set min/max word distance.
        wordDistanceRegex = '{' + str(min_nbr_words) + ',' + str(max_nbr_words) + '}'

        if rs1.optional:
            ending_paren_pos = -1
            if rs1.whole_word:
                # Things get a little messy if both rs1.optional and rs1.whole_word.
                ending_paren_pos = -3
                # We don't have to worry about the \b for rs1 because
                # `interveningPuncAndSpace` contains a \b.

            # (Assuming that it ends in '?'.)
            # Find the left parens which the last '?' pertains to
            leftParenIdx = rs1.get_regex_str().rfind('(')

            if min_nbr_words == 0 and max_nbr_words == 0:
                firstPart = rs1.get_regex_str()[0:leftParenIdx] + \
                            '(?:' + rs1.get_regex_str()[leftParenIdx : ending_paren_pos] + \
                            interveningPuncAndSpace + \
                            ')?'
                joinStrRegex = ''
            else:
                # Not sure about this
                firstPart = rs1.get_regex_str()[0:leftParenIdx] + \
                            '(?:' + rs1.get_regex_str()[leftParenIdx:-1] + \
                            interveningPunc + ')?'
                joinStrRegex = readyForDistanceRangeWordRegex + wordDistanceRegex + '\s'
        else:
            firstPart = rs1.get_regex_str() + interveningPunc

            if min_nbr_words == 0 and max_nbr_words == 0:
                joinStrRegex = '\s'
            else:
                joinStrRegex = readyForDistanceRangeWordRegex + wordDistanceRegex + '\s'


        # Create an empty/invalid RegexString object, and make it valid by editing
        # its regex_str property.
        resultRegexString = RegexString([''])
        resultRegexString.regex_str = firstPart + joinStrRegex + rs2.regex_str

        if rs2.optional:
            resultRegexString.optional = True
        else:
            resultRegexString.optional = False

        return resultRegexString


    @staticmethod
    def regex_to_RegexString(regex: str) -> 'RegexString':
        """
        Build a RegexString object from the given regular
        expression.
        Args:
            regex (str): 

        Returns:
            RegexString: 
        """
        # Create an empty/invalid RegexString object, and make it valid by editing
        # its regex_str property.
        result = RegexString([''])
        result.regex_str = regex
        return result


    @staticmethod
    def build_regex_string(input_list: List[str]) -> Self:
        """
        Read the input list to create a new RegexString object from 
        the concatenated elements.

        Args:
            input_list (List[str]): a list having an odd number of elements 
            following the pattern "phraseList1, maxDistance1, phraseList2, 
            maxDistance2, phraseList3, ...", each phraseList consisting of a 
            list of strings and each distance consisting of an integer >= 0;
            each phraseList will be used to create a new RegexString object, 
            after which each RegexString distance RegexString triple will be 
            concatenated to create another set of RegexString objects

        Raises:
            ValueError: if the input list has an invalid number of elements

        Returns:
            RegexString: a single RegexString object concatenating all the 
                list items
        """
        if len(input_list) < 3 or len(input_list) % 2 == 0:
            msg = "Input parameter list must have an odd number of elements "
            msg += "following this pattern:\n"
            msg += "   'phraseList1, maxDistance1, phraseList2, maxDistance2, "
            msg += "phraseList3, ...'"
            raise ValueError(msg)

        finalRegexStr = None
        currRegexStr = RegexString(input_list[0])
        currMaxDistance = input_list[1]
        idx = 2
        while idx < len(input_list):
            nextRegexStr = RegexString(input_list[idx])
            finalRegexStr = RegexString.concat_with_word_distances(currRegexStr, 
                                                                   nextRegexStr, 
                                                                   max_nbr_words=currMaxDistance)
            if idx + 2 < len(input_list):
                currRegexStr = finalRegexStr
                currMaxDistance = input_list[idx+1]
                idx += 2
            else:
                idx += 1

        return finalRegexStr


    def __repr__(self):
        return self.regex_str


    def __eq__(self, other):
        if type(self) != type(other):
            return False
        elif self is other:
            return True
        elif self.get_regex_str() == other.get_regex_str():
            return True
        return False


if __name__ == '__main__':
    import inspect

    # Stamp descriptions from https://www.mysticstamp.com/.

    input = \
    """
    # 11A - 1853-55 3¢ George Washington, dull red, type II, imperf

    # 17 - 1851 12c Washington imperforate, black

    # 12 - 1856 5c Jefferson, red brown, type I, imperforate

    # 18 - 1861 1c Franklin, type I, perf 15

    # 40 - 1875 1c Franklin, bright blue

    # 42 - 1875 5c Jefferson, orange brown

    # 62B - 1861 10c Washington, dark green
    """
    input = inspect.cleandoc(input)

    type_markers = ['type', 'Type']
    type_rs = RegexString(type_markers, whole_word=True)
    print(f"Regular expression for 'type_rs': {type_rs.get_regex_str()}\n")

    roman_numerals = ['I', 'II', 'III', 'IV', 'V']
    roman_nums_rs = RegexString(roman_numerals, whole_word=True)
    print(f"Regular expression for 'roman_nums_rs': {roman_nums_rs.get_regex_str()}")

    type_phrase_rs = RegexString.concat_with_word_distances(type_rs, 
                                                            roman_nums_rs,
                                                            min_nbr_words=0, 
                                                            max_nbr_words=0)
    matchStrs = type_phrase_rs.get_match_triples(input)
    print("\nType info found in the input:")
    print(matchStrs)

    cent_symbols = ['c', '¢']
    # Use `prepend` to look for one or two digits before the cent symbol.
    cent_rs = RegexString(cent_symbols, prepend='\d\d?')
    print(f"\nRegular expression for 'cent_rs': {cent_rs.get_regex_str()}")
    matchStrs = cent_rs.get_match_triples(input)
    print("\nCents info found in the input:")
    print(matchStrs)

    perforation_markers = ['perf', 'imperforate', 'imperf']
    perforations_rs = RegexString(perforation_markers)
    matchStrs = perforations_rs.get_match_triples(input)
    print("\nPerforation info:")
    print(matchStrs)

    # But if the stamp is perforated we want to also extract the number which follows.
    imperforated_markers = ['imperforate', 'imperf']
    imperf_rs = RegexString(imperforated_markers)
    matchStrs = imperf_rs.get_match_triples(input)
    print("\nImperforated:")
    print(matchStrs)

    perforation_markers = ['perf']
    perf_rs = RegexString(perforation_markers, append=' \d\d')
    matchStrs = perf_rs.get_match_triples(input)
    print("\nPerforated sizes:")
    print(matchStrs)

    colors = ['black', 'blue', 'brown', 'green', 'orange', 'red']

    colors_rs = RegexString(colors, whole_word=True)
    matchStrs = colors_rs.get_match_triples(input)
    print("\nColors alone:")
    print(matchStrs)

    # Create the qualifiers.
    # Add colors to the qualifiers, so that we can ID, e.g. 'orange brown'.
    color_qualifiers = ['bright', 'dark', 'dull'] + colors

    inputList = [color_qualifiers, 0, colors]
    color_phrase_rs = RegexString.build_regex_string(inputList)
    matchStrs = color_phrase_rs.get_match_triples(input)
    print("\nColor qualifiers + colors:")
    print(matchStrs)

    # Not getting the colors not preceded by a qualifier.
    # Make the qualifier optional.
    color_qualifiers_rs = RegexString(color_qualifiers, 
                                      whole_word=True, 
                                      optional=True
                                    )
    color_phrase_rs = RegexString.concat_with_word_distances(color_qualifiers_rs, 
                                                            colors_rs,
                                                            min_nbr_words=0, 
                                                            max_nbr_words=0)
    matchStrs = color_phrase_rs.get_match_triples(input)
    print("\nOptional color qualifiers + colors found in the input:")
    print(matchStrs)


    id_symbols = ['#']

    id_rs = RegexString(id_symbols, append='\s\d+')
    matchStrs = id_rs.get_match_triples(input)
    print("\nStamp IDs:")
    print(matchStrs)

    # Not getting the optional letter.
    id_rs = RegexString(id_symbols, append='\s\d+(?:\w+)?')
    matchStrs = id_rs.get_match_triples(input)
    print("\nStamp IDs with letters found in the input:")
    print(matchStrs)



from text_to_relations.relation_extraction import StringUtils
from text_to_relations.relation_extraction import SpacyUtils
from text_to_relations.relation_extraction.Annotation import Annotation
from typing import Tuple, List, Self

class TokenAnn(Annotation):
    # A class of objects representing word tokens which "know" their 
    # starting and ending offsets in a source document. 
    
    # Contractions and the possessive 's are considered word tokens despite the apostrophe punctuation 
    # which they contain.
    kindExceptions = ["'s", "'ve", "'d", "'ll", "n't"]

    def __init__(self, start_offset, end_offset, contents):
        """"""
        if contents in TokenAnn.kindExceptions:
            kind = 'word'
        elif StringUtils.isAllPunc(contents):
            kind = 'punc'
        elif StringUtils.isAllWordChars(contents):
            kind = 'word'
        else: 
            kind = 'other'
        
        features = {'kind': kind}
        super().__init__('Token', contents, start_offset, end_offset, features)


    @staticmethod
    def build_annotation_distance_regex(first_ann: Annotation, 
                                        word_distance_range: Tuple[int, int], 
                                        token_kind: str, 
                                        second_ann: Annotation) -> str:
        """
        Build a string regular expression that specifies the token distance between two annotations
        necessary for a match.

        Args:
            first_ann (Annotation): 
            word_distance_range (Tuple[int, int]): a pair of integers, whose first element is a minimum token distance
        and whose second element is a maximum token distance 
            token_kind (str): the Token.kind property necessary for a match; None if all Token objects
        are to match
            second_ann (Annotation): 

        Returns:
            str: a regular expression
        """
        """
        :param first_ann: 
        :param word_distance_range: 
        :param token_kind: 
        :param second_ann: 
        :return: 
        """
        distanceToken = 'Token'

        result = r"<'"
        result += first_ann
        result += r"[^>]*>(?:<'"
        result += distanceToken

        if token_kind is None:
            result += r"[^>]*>){"
        else:
            result += r"'[^>]*kind='"
            result += token_kind
            result += r"'[^>]*>){"

        minTs, maxTs = word_distance_range
        result += str(minTs) + ',' + str(maxTs)
        result += r"}<'"
        result += second_ann
        result += r"[^>]*>"

        return result


    @staticmethod
    def get_token_objects(input_str: str, 
                          start_pos_in_doc: int) -> List[Self]:
        """
        Create TokenAnn objects on the given substring of a longer document.
        Args:
            input_str (str): a substring of some document
            start_pos_in_doc (int): the starting offset of inputStr in the 
                source document

        Returns:
            List[Self]: a list of TokenAnn objects
        """
        result = []

        # This is how we track the starting and ending position of each token. 
        # For each token, we find the first instance in this string--and use
        # that to determine the starting position. We replace the found substring
        # with x's to avoid the problem of repeating tokens.
        consumedStr = input_str

        tokenStrs = SpacyUtils.tokenize(input_str)

        # lastPos = 0
        for tokenStr in tokenStrs:
            startPosInInput = consumedStr.index(tokenStr)
            xStr = 'x' * len(tokenStr)
            consumedStr = consumedStr.replace(tokenStr, xStr, 1)
            endPosInInput = startPosInInput + len(tokenStr)
            token = TokenAnn(start_pos_in_doc + startPosInInput, start_pos_in_doc + endPosInInput, tokenStr)
            result.append(token)

        return result
    
    
    @staticmethod
    def text_to_token_anns(text_input: str) -> List[Self]:
        """
        Split the given input text into tokens, and create a TokenAnn
        on each one.

        Args:
            text_input (str): 

        Returns:
            List[Self]: a list of TokenAnn annotations created on the given text
        """
        tokenStrs = SpacyUtils.tokenize(text_input)

        result = []
        startSearchIdx = 0
        for token in tokenStrs:
            startIdx = text_input.find(token, startSearchIdx)
            endIdx = startIdx + len(token)
            tokenAnn = TokenAnn(startIdx, endIdx, token.strip())
            result.append(tokenAnn)
            startSearchIdx = endIdx

        return result


if __name__ == '__main__':
    pass



import re
from typing import Dict, Self, Collection, List

from text_to_relations.relation_extraction import StringUtils

class Annotation(object):
    """"""
    regexQuote = r"['\"].*?['\"]"

    def __init__(self, type: str, contents: str, 
                 start_offset: int, end_offset: int, 
                 properties: Dict[str, object]=None):
        """

        Args:
            type (str): annotation (entity) type, e.g. Person, Currency, River
            contents (str): the mention in the doc, e.g. 'John Smith', 'Euros', 'Amazon'
            start_offset (int): 
            end_offset (int): 
            properties (Dict[str, object], optional): A free-form dict for 
                adding attributes. Defaults to None.
        """
        if start_offset > end_offset:
            msg = "Start offset cannot be greater than end offset. Start: " + str(start_offset) + "; End: " + str(end_offset)
            raise ValueError(msg)

        if start_offset < 0 or end_offset < 0:
            msg = "Start and end offset cannot be less than 0. Start: " + str(start_offset) + "; End: " + str(end_offset)
            raise ValueError(msg)

        self.type = type
        self.start_offset = start_offset
        self.end_offset = end_offset

        cleanedContents = contents.replace('\n', ' ')
        self.normalizedContents = StringUtils.removeMultipleSpaces(cleanedContents).strip()

        if properties is None:
            properties = {}
        self.properties = properties
    

    def to_dict(self) -> Dict[str, str]:
        result = {'type': self.type, 
                  'start': self.start_offset, 
                  'end': self.end_offset, 
                  'text': self.normalizedContents}
        return result


    def __repr__(self):
        if self.properties == {}:
            result = "<'%s'(normalizedContents='%s', start='%s', end='%s')>" \
                     % (self.type, self.normalizedContents, self.start_offset, self.end_offset)
        else:
            features = ''
            for featureName in self.properties:
                features += featureName + "='" + self.properties[featureName] + "', "
            # Remove last comma-space.
            features = features[0:-2]
                
            result = "<'%s'(normalizedContents='%s', start='%s', end='%s', %s)>" \
                     % (self.type, self.normalizedContents, self.start_offset, self.end_offset, features)

        return result

    def __eq__(self, other):
        if not isinstance(other, Annotation):
            return False
        elif self is other:
            return True
        else:
            if self.properties != other.properties:
                return False
            
            if self.type == other.type and \
                        self.start_offset == other.start_offset and \
                        self.end_offset == other.end_offset and \
                        self.normalizedContents == other.normalizedContents:
                    return True

        return False
    
    def __hash__(self):
        return hash(self.__repr__())

    @staticmethod
    def sort(items: Collection[Self]) -> Collection[Self]:
        """
        Sort the in-coming collection of Annotations by starting and
        ending offset.
        Args:
            items (Collection[Self]): 

        Returns:
            Collection[Self]: 
        """
        anns = sorted(items, key=lambda p: (p.start_offset, p.end_offset), reverse=False)
        return anns

    @staticmethod
    def str_to_Annotation(annStr: str) -> Self:
        """
        Convert the output of __repr__ to an Annotation object.
        Assuming that the strings are in one of these two forms:
        - "<'AnnotationName'(normalizedContents='...', start='m', end='n')>"
        - '<"AnnotationName"(normalizedContents="...", start="m", end="n")>'

        Args:
            annStr (str): 

        Returns:
            Self: 
        """
        # TODO: Currently this method will not add any unexpected strings (i.e. features) into the properties.
        matches = re.findall(Annotation.regexQuote, annStr)
        # We use '[1:-1]' to chop off the quote at either end.
        aType = matches[0][1:-1]
        contents = matches[1][1:-1]
        start = int(matches[2][1:-1])
        end = int(matches[3][1:-1])
        ann = Annotation(aType, contents, start, end)
        return ann
    
    
    @staticmethod
    def encloses(ann1: Self, ann2: Self) -> bool:
        """
        Does ann1 "enclose" ann2? I.e., is ann2 entirely contained within the bounds of ann1?
        Determination is made with starting and ending offsets only--the annotations'
        normalizedContents properties are ignored.
        Args:
            ann1 (Self): 
            ann2 (Self): 

        Returns:
            bool: 
        """
        if ann1.start_offset <= ann2.start_offset and ann1.end_offset >= ann2.end_offset:
            return True
        return False
    
    
    @staticmethod
    def get_enclosed(ann: Self, ann_list: List[Self]) -> List[Self]:
        """
        Return the subset of the annotations in the given list which are enclosed
        by the given annotation.
        Args:
            ann (Self): 
            ann_list (List[Self]): 

        Returns:
            List[Self]: 
        """
        result = []
        for annElement in ann_list:
            if Annotation.encloses(ann, annElement):
                result.append(annElement)
        
        return result
    

if __name__ == '__main__':
    pass
